import json

message_content_types = {
    'text',
    'audio',
    'animation',
    'document',
    'game',
    'photo',
    'sticker',
    'contact'
}

class Message:
    def __init__(self, source:dict):
        self.message_id = source['message_id']
        self.content_type = None
        self.command = None
        self.contact = None
        self.audio = None
        self.animation = None
        self.photo = None
        self.sticker = None
        self.game = None
        self.text = ''
        if 'text' in source:                
            self.text:str = source['text']
            self.content_type='text'
            if self.text.startswith('/'):
                # this is a comand
                command = self.text.split()[0].split('@')[0][1:] 
                self.command = command
        if 'audio' in source:
            self.content_type='audio'
            self.audio = source['audio']
        if 'animation' in source:
            self.content_type='animation'
            self.animation = source['animation']
        if 'document' in source:
            self.content_type='document'
            self.document = source['document']
        if 'game' in source:
            self.content_type='game'
            self.game = source['game']
        if 'photo' in source:
            self.content_type='photo'
            self.photo = source['photo']
        if 'sticker' in source:
            self.content_type='sticker'
            self.sticker = source['sticker']
        if 'contact' in source:
            self.content_type='contact'
            self.contact = source['contact']

class ReplyKeyboard:
    def __init__(self):
        self.rows=[[]]
        self.remove_keyboard = False
        self.resize_keyboard = False
        self.one_time_keyboard = False
        pass
    
    def clear(self):
        self.rows = [[]]

    def add_button(self, name, row, request_contact=False,request_location=False):
        button = {
            'text' : name,
            'request_contact' : request_contact,
            'request_location' : request_location
        }
        self.rows[row].append(button)
        pass

    def add_row(self):
        self.rows.append([])
        pass

    def to_json(self)->str:
        res = {
            'keyboard' : self.rows,
            'resize_keyboard' : True,
            'one_time_keyboard' : True,
            'remove_keyboard' : self.remove_keyboard
        }
        return json.dumps(res)



from vk_py_telegrambot import api
from vk_py_telegrambot import store
from vk_py_telegrambot import types
from vk_py_telegrambot import users

import re
import logging
logger = logging.getLogger('bot')

import time

def default_finaly_handler(bot, message, user):
    bot.reply(user=user,message={'text':'Прости, я не понимаю :)'})

def _test(_filter, message):

    if not message.content_type in _filter['content_types']:
        return False

    if 'commands' in _filter:
        if message.command in _filter['commands']: return True
    
    if 're_text' in _filter:
        if len(re.findall(_filter['re_text'],message.text))>0:
            return True
    if 'func' in  _filter:
        return _filter['func'](message)

    return False

class BotState:
    name : str
    state_set_handler : callable
    state_fadeout_handler : callable
    finaly_handler : callable
    _handlers : list

    def __init__(self, name):
        self.name = name
        self._handlers=[]

    
    def addHandler(self, handler, content_types=['text'], commands=None, func=None, re_text=None):
        _filter = { 'content_types' : content_types}
        if not commands is None: _filter['commands']=commands
        if not re_text is None: _filter['re_text']=re_text
        if not func is None: _filter['func']=func
        
        self._handlers.append((_filter, handler))
    
    def proceed_message(self, bot, message, user):
        for record in self._handlers:
            if not _test(record[0], message):
                continue
            result = False
            try:
                result = record[1](bot, message, user)
            except Exception as e:
                logger.info(e)
            if result:
                break
        
        if self.finaly_handler:
            self.finaly_handler(bot, message, user)
        else:
            logger.info('Состояние бота: {}, Не обработано сообщение {} от пользователя {}'.format(self.name, message, user))

    def on_state_set(self, bot, user):
        if self.state_set_handler:
            try:
                self.state_set_handler(bot, user)
            except Exception as e:
                logger.info(e)

    def on_state_fadeout(self, bot, user):
        if self.state_fadeout_handler:
            try:
                self.state_fadeout_handler(bot, user)
            except Exception as e:
                logger.info(e)

default_state = BotState('default')
default_state.finaly_handler = default_finaly_handler

class Bot:

    def __init__(self, token:str, proxy={}, skip_pending:bool=False, num_threads:int=2):
        self.token = token
        self.store = store.SessionStore()
        self.last_update_id = 0
        self.skip_pending = skip_pending
        self.default_state = default_state
        self.states:dict = {} 
        self.mw_handlers = []   

        if not proxy is None:api.proxy=proxy

        self.jobs_queue = None
        self.running = False


    def add_state(self, state:BotState):
        if state.name in self.states:
            logger.warn('Состояние переопределяется {}. возможно это проблема'.format(state.name))

        self.states[state.name] = state
    
    def _send_pending_messages(self):
    # Забираем отложенные сообщения и отправляем пользователям
        pending_messages = self.store.get_pending_messages()
        for message in pending_messages:
            print(message)
            pass
    def _check_for_state_events(self):
    # Смотрим в базе по каким пользователям наступило время сбрасывать статус    
        events = self.store.get_chats_with_expired_stages()
        for event in events:
            print(event)
            pass

    def _proceed_updates(self):
    # Обрабатываем обновления от бота
        updates = api.get_updates(self.token,offset=self.last_update_id+1)
        for ju in updates:
            self.store.save_message(ju)
            if 'message' in ju: 
                #разбираем что пришло
                message = types.Message(ju['message'])
                user = users.User(source=ju['message']['from'], store=self.store)
                #вызываем промежуточные обработчики
                for mw in self.mw_handlers:
                    mw(bot=self, message=message, user=user)

                #понимаем текущее состояние и обрабатываем по состоянию
                if user.state in self.states:
                    self.states[user.state].proceed_message(bot=self, message=message, user=user)
                else:
                    self.default_state.proceed_message(bot=self, message=message, user=user)

            self.last_update_id=ju['update_id']

    def start(self, none_stop=False, interval=0, timeout=20):
        logger.info('Started .......')
        self.running=True
        while self.running:
            try:
                logger.debug('Pull {}'.format(self.last_update_id))
                
                self._send_pending_messages()

                self._check_for_state_events()

                self._proceed_updates()

                time.sleep(timeout)
            except KeyboardInterrupt:
                logger.info("KeyboardInterrupt received.")
                break
        logger.info('Stopped polling.')

    def stop(self):
        self.running=False
        logger.info("Bot was stopped")
        pass

    def send_message(self, user, message, when=None, new_state=None,  reply_markup=None):
        """
            :when: - Дата и время, когда сообщение должно быть отправлено. если не заполнено отправляется немеделенно
            :new_state: - состоения в которое должен перейти чат после отправки
            :reply_markup: - клавиатура для ответа. Можно собрать руками, но проще используя класс  ReplyKeyboard
        """
        if when is None:
            # мы отправляем сообщение немедленно
            if 'text' in message:
                api.send_message(
                    token=self.token,
                    chat_id=user.chat_id,
                    text = message['text'],
                    reply_markup=reply_markup,
                )
                pass
            if 'photo' in message:
                pass
        else:
            # мы сохраняем сообщение в базе
            self.store.put_message_in_queue(
                chat_id=user.chat_id,
                message=message,
                when=when, 
                new_state=new_state,
                reply_markup=reply_markup,
            )

    def reply(self, user, message, reply_markup=None, new_state=None):
        if 'text' in message:
            api.send_message(
                    token=self.token,
                    chat_id=user.chat_id,
                    text = message['text'],
                    reply_markup=reply_markup
            )
        
        if not new_state is None: 
            if not new_state[0] in self.states: 
                logger.warn('Для пользователя {}({}) установлено состояние {} для которого нет описания. Это может быть проблемой'.format(user.name, user.chat_id, new_state[0]))
            else:
                self.states[new_state[0]].on_state_set(bot=self, user=user)
            
            user.set_new_state(new_state)


import json
import logging

logger = logging.getLogger('chatbot')

try:
    import requests
except ImportError:
    logger.error("can't import requests module")
    exit(500)

try:    
    from requests.packages.urllib3 import fields
    format_header_param = fields.format_header_param
except ImportError:
    format_header_param = None

proxy = None

API_URL = None
FILE_URL = None

CONNECT_TIMEOUT = 3.5
READ_TIMEOUT = 9999

class APIException(Exception):

    def __init__(self, msg, function_name, result):
        super(APIException, self).__init__("A request to the Telegram API was unsuccessful. {0}".format(msg))
        self.function_name = function_name
        self.result = result

def _get_req_session(reset=False):
    session = requests.Session()
    return session

def _make_request(token, method_name, method='get', params=None, files=None):
    """
    Makes a request to the Telegram API.
    :param token: The bot's API token. (Created with @BotFather)
    :param method_name: Name of the API method to be called. (E.g. 'getUpdates')
    :param method: HTTP method to be used. Defaults to 'get'.
    :param params: Optional parameters. Should be a dictionary with key-value pairs.
    :param files: Optional files.
    :return: The result parsed to a JSON dictionary.
    """
    if API_URL is None:
        request_url = "https://api.telegram.org/bot{0}/{1}".format(token, method_name)
    else:
        request_url = API_URL.format(token, method_name)
    
    logger.debug("Request: method={0} url={1} params={2} files={3}".format(method, request_url, params, files))
    read_timeout = READ_TIMEOUT
    connect_timeout = CONNECT_TIMEOUT
    if files and format_header_param:
        fields.format_header_param = _no_encode(format_header_param)
    if params:
        if 'timeout' in params: read_timeout = params['timeout'] + 10
        if 'connect-timeout' in params: connect_timeout = params['connect-timeout'] + 10
    result = _get_req_session().request(method, request_url, params=params, files=files,
                                        timeout=(connect_timeout, read_timeout), proxies=proxy)
    logger.debug("The server returned: '{0}'".format(result.text.encode('utf8')))
    return _check_result(method_name, result)['result']

def _check_result(method_name, result):
    """
    Checks whether `result` is a valid API response.
    A result is considered invalid if:
        - The server returned an HTTP response code other than 200
        - The content of the result is invalid JSON.
        - The method call was unsuccessful (The JSON 'ok' field equals False)

    :raises ApiException: if one of the above listed cases is applicable
    :param method_name: The name of the method called
    :param result: The returned result of the method request
    :return: The result parsed to a JSON dictionary.
    """
    if result.status_code != 200:
        msg = 'The server returned HTTP {0} {1}. Response body:\n[{2}]' \
            .format(result.status_code, result.reason, result.text.encode('utf8'))
        raise APIException(msg, method_name, result)

    try:
        result_json = result.json()
    except:
        msg = 'The server returned an invalid JSON response. Response body:\n[{0}]' \
            .format(result.text.encode('utf8'))
        raise APIException(msg, method_name, result)

    if not result_json['ok']:
        msg = 'Error code: {0} Description: {1}' \
            .format(result_json['error_code'], result_json['description'])
        raise APIException(msg, method_name, result)
    return result_json

def get_me(token):
    method_url = r'getMe'
    return _make_request(token, method_url)

def send_message(token, chat_id, text, disable_web_page_preview=None, reply_to_message_id=None, reply_markup=None,
                 parse_mode=None, disable_notification=None, timeout=None, *args, **kwargs):
    """
    Use this method to send text messages. On success, the sent Message is returned.
    :param token:
    :param chat_id:
    :param text:
    :param disable_web_page_preview:
    :param reply_to_message_id:
    :param reply_markup:
    :param parse_mode:
    :param disable_notification:
    :return:
    """
    method_url = r'sendMessage'
    payload = {'chat_id': str(chat_id), 'text': text}
    if disable_web_page_preview:
        payload['disable_web_page_preview'] = disable_web_page_preview
    if reply_to_message_id:
        payload['reply_to_message_id'] = reply_to_message_id
    if reply_markup:
        payload['reply_markup'] = reply_markup
    if parse_mode:
        payload['parse_mode'] = parse_mode
    if disable_notification:
        payload['disable_notification'] = disable_notification
    if timeout:
        payload['connect-timeout'] = timeout
    return _make_request(token, method_url, params=payload, method='post')

def send_photo(token, chat_id, photo, caption=None, reply_to_message_id=None, reply_markup=None,
               parse_mode=None, disable_notification=None):
    method_url = r'sendPhoto'
    payload = {'chat_id': chat_id}
    files = None
    if not isinstance(photo, str):
        files = {'photo': photo}
    else:
        payload['photo'] = photo
    if caption:
        payload['caption'] = caption
    if reply_to_message_id:
        payload['reply_to_message_id'] = reply_to_message_id
    if reply_markup:
        payload['reply_markup'] = _convert_markup(reply_markup)
    if parse_mode:
        payload['parse_mode'] = parse_mode
    if disable_notification:
        payload['disable_notification'] = disable_notification
    return _make_request(token, method_url, params=payload, files=files, method='post')


def get_updates(token, offset=None, limit=None, timeout=None, allowed_updates=None):
    method_url = r'getUpdates'
    payload = {}
    if offset:
        payload['offset'] = offset
    if limit:
        payload['limit'] = limit
    if timeout:
        payload['timeout'] = timeout
    if allowed_updates:
        payload['allowed_updates'] = json.dumps(allowed_updates)
    return _make_request(token, method_url, params=payload)

def answer_callback_query(token, callback_query_id, text=None, show_alert=None, url=None, cache_time=None):
    """
    Use this method to send answers to callback queries sent from inline keyboards. The answer will be displayed to the user as a notification at the top of the chat screen or as an alert. On success, True is returned.
    Alternatively, the user can be redirected to the specified Game URL. For this option to work, you must first create a game for your bot via BotFather and accept the terms. Otherwise, you may use links like telegram.me/your_bot?start=XXXX that open your bot with a parameter.
    :param token: Bot's token (you don't need to fill this)
    :param callback_query_id: Unique identifier for the query to be answered
    :param text: (Optional) Text of the notification. If not specified, nothing will be shown to the user, 0-200 characters
    :param show_alert: (Optional) If true, an alert will be shown by the client instead of a notification at the top of the chat screen. Defaults to false.
    :param url: (Optional) URL that will be opened by the user's client. If you have created a Game and accepted the conditions via @Botfather, specify the URL that opens your game – note that this will only work if the query comes from a callback_game button.
    Otherwise, you may use links like telegram.me/your_bot?start=XXXX that open your bot with a parameter.
    :param cache_time: (Optional) The maximum amount of time in seconds that the result of the callback query may be cached client-side. Telegram apps will support caching starting in version 3.14. Defaults to 0.
    :return:
    """
    method_url = 'answerCallbackQuery'
    payload = {'callback_query_id': callback_query_id}
    if text:
        payload['text'] = text
    if show_alert:
        payload['show_alert'] = show_alert
    if url:
        payload['url'] = url
    if cache_time is not None:
        payload['cache_time'] = cache_time
    return _make_request(token, method_url, params=payload, method='post')


def answer_inline_query(token, inline_query_id, results, cache_time=None, is_personal=None, next_offset=None,
                        switch_pm_text=None, switch_pm_parameter=None):
    method_url = 'answerInlineQuery'
    payload = {'inline_query_id': inline_query_id, 'results': _convert_list_json_serializable(results)}
    if cache_time is not None:
        payload['cache_time'] = cache_time
    if is_personal:
        payload['is_personal'] = is_personal
    if next_offset is not None:
        payload['next_offset'] = next_offset
    if switch_pm_text:
        payload['switch_pm_text'] = switch_pm_text
    if switch_pm_parameter:
        payload['switch_pm_parameter'] = switch_pm_parameter
    return _make_request(token, method_url, params=payload, method='post')

def _convert_markup(markup):
    # if isinstance(markup, types.JsonSerializable):
    #     return markup.to_json()
    # return markup
    return []

def _convert_list_json_serializable(results):
    # ret = ''
    # for r in results:
    #     if isinstance(r, types.JsonSerializable):
    #         ret = ret + r.to_json() + ','
    # if len(ret) > 0:
    #     ret = ret[:-1]
    # return '[' + ret + ']'
    return []

def _no_encode(func):
    def wrapper(key, val):
        if key == 'filename':
            return u'{0}={1}'.format(key, val)
        else:
            return func(key, val)

    return wrapper



import logging
import os
import re
import sys
import time


logger = logging.getLogger('bot')
logger.setLevel(logging.DEBUG)
formatter = logging.Formatter(
    '%(asctime)s (%(filename)s:%(lineno)d %(threadName)s) %(levelname)s - %(name)s: "%(message)s"'
)

console_output_handler = logging.StreamHandler(sys.stderr)
console_output_handler.setFormatter(formatter)
logger.addHandler(console_output_handler)

from vk_py_telegrambot import bot



from vk_py_telegrambot import store

class User:
    chat_id:str
    name: str
    state: str
    context : dict
    store : store.SessionStore
    data : dict

    def __init__(self, source:dict, store:store.SessionStore):
        self.chat_id = source['id']
        if 'username' in source:
            self.name = source['username']
        else:
            self.name = source['first_name']
        self.store = store

        self.load()

    def load (self):
        ud = self.store.get_userdata_by_id(self.chat_id) 
        if 'context' in ud:
            self.context = ud['context']
        if 'state' in ud:
            self.state = ud['state']

    def save(self, store):
        pass

    def set_state(self, state):
        self.store.set_new_state(self.chat_id, state[0], state[1])
    
    def get_attr(self, attr):
        if attr in self.context: 
            return self.context[attr]
        else:
            return None
    
    def set_attr(self, attr, value):
        self.context[attr]=value

        if not self.store is None:
            self.store.set_locals_by_id(self.chat_id, self.context)

    def save_phone(self, phone_number):
        self.store.save_phone(chat_id=self.chat_id, phone=phone_number)



import logging
from datetime import datetime
from datetime import timedelta
logger = logging.getLogger('chatbot')

try:
    import pymongo
except ImportError:
    logger.error("Can't import pymongo module. Check requirements.txt")
    exit(500)

from datetime import datetime 

"""
    SessionStore - хранилище данных для бота
"""
class SessionStore:
    def __init__(self):
        self.conn = pymongo.MongoClient()       
        self.db = self.conn['bot-session-store']
        self.sessions = self.db['sessions']
        self.pending_queue = self.db['pending_queue']
        logger.debug('Connection to mongoclinet estabished. We find {} sessions'.format(self.sessions.find().count()))

    def save_message(self, message):
        self.db['messages'].insert_one(message)
        
    def get_userdata_by_id(self, chat_id)->dict:
        """Получает значение переменных для чата"""
        
        user_data = {}
        try:
            query = self.sessions.find_one({"chat_id":chat_id})
            if query:
                logger.debug('Query to db with result {}'.format(query))
                user_data = query
            return user_data
        except Exception as identifier:
            print(identifier)        

        return user_data

    def set_locals_by_id(self, chat_id, data):
        """Устанавливает значение переменных для чата"""
        filter = {"chat_id"   : chat_id}
        new_doc = {
            "chat_id"       : chat_id,
            "context"       : data,
            "stage"         : None,
            "stage_expired" : None
        }

        rec = self.sessions.find_one(filter=filter)

        if rec == None: 
            self.sessions.insert_one(document=new_doc)
        else:
            new_doc['stage']= rec['stage']
            new_doc['stage_expired']= rec['stage_expired']

            self.sessions.replace_one(filter=filter, replacement=new_doc)
        pass

    def set_new_state(self, chat_id:str, new_state:str, ttl:int):
        """Переводит чат в стадию. Стадия используется для контекстной переписки"""
        if ttl>0:
            new_state_ttl = datetime.now() + timedelta(seconds=ttl)
        else:
            new_state_ttl = datetime.now() + timedelta(days=1)
            
        query = {'chat_id':chat_id}
        record = self.sessions.find_one(query)
        if record is None:
           record = {
               'chat_id' : chat_id,
               'state' : new_state,
               'state_ttl' : new_state_ttl
           } 
           self.sessions.insert_one(record)
        else:
            self.sessions.find_one_and_update(
                filter=query,update={
                    '$set':{
                        'state':new_state, 
                        'state_ttl': new_state_ttl
                }})
             
        logger.debug('Для пользователя {} установлено новое состояние {} на срок {} секунд'.format(chat_id, new_state,ttl))
        
        pass

    def drop_state(self, chat_id:str):
        """Сбрасывает текущую стадию с чата"""
        logger.debug('Для пользователя {} состояние сброшено'.format(chat_id))
        pass

    def get_chats_with_expired_stages(self):
        exp_sessions = self.sessions.find({'state_ttl':{'$lte':datetime.now()}})
        if exp_sessions.count()==0:
            return []
        else:
            logger.debug('Нашли {} стадии время ожидания по которым истекло'.format(exp_sessions.count()))
            events = []
            for s in exp_sessions:
                logger.debug('  chat_id: {}, state: {}, ttl: {}'.format(s['chat_id'], s['state'], s['state_ttl']))
                event = {
                    'event':'state_expired',
                    'state':s['state'],
                    'chat_id':s['chat_id']
                }
                if 'context' in s:
                    event['context'] = s['context']
                
                events.append(event)
                self.sessions.find_one_and_update(
                    filter={'_id':s['_id']},
                    update={'$set':{
                        'state':None,
                        'state_ttl':None
                }})
            
            return events
        pass

    def get_pending_messages(self)->[dict]:
        query = {
            'fire_date':{ '$lte':datetime.now()},
            'is_active':True
        }
        update = {'$set':{'is_active':False}}

        queue = self.pending_queue.find_and_modify(query=query, update=update)
        
        if queue is None:
            queue=[]

        if type(queue)==dict: 
            return [queue]

        return queue

    def put_message_in_queue(self, chat_id, message, when, new_state=None, reply_markup=None):
        if when < datetime.now():
            logger.warn('Попытка отправить отложенное сообщение которое уже просроченно')

        el = {
            'message' : message,
            'fire_date' : when,
            'chat_id' : chat_id,
            'new_state' : new_state[0],
            'new_state_ttl' : new_state[1],
            'is_active' : True,
            'reply_markup' : reply_markup,
        }
        self.pending_queue.insert_one(el)

    def drop_session(self, id):
        self.sessions.delete_one({'chat_id':id})

    def __del__(self):
        if self.db:
            self.conn.close()
            self.conn = None
            self.db = None
            logger.debug('mongoclient was shuting down....')

    def save_phone(self, chat_id, phone):
        ud = self.sessions.find_one({'chat_id' : chat_id})
        if not ud:
            self.sessions.insert_one({
                "chat_id"       : chat_id,
                "context"       : {},
                "stage"         : None,
                "stage_expired" : None,
                "phone_number"  : phone         
            })
        else:
            self.sessions.find_one_and_update(
                filter={'chat_id':chat_id},
                update={'$set':{'phone':phone}}
            )